import sys
import logging
import datetime
import json
import asyncio
from dataclasses import dataclass

from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.responses import Response, HTMLResponse

import uvicorn

import zmq
from zmq.asyncio import Context as ZmqContext
from zmq.asyncio import Socket as ZmqSocket

from package.shared import models
from package.shared.logger import setup_logger
from package.shared.config import Config
from package.shared.ipc import IPCWorker
from package.shared.util import board_cfg_file


@dataclass
class IPCImpl(IPCWorker):
    task: asyncio.Task = None
    m_ctx: ZmqContext = None


    def __post_init__(self):
        super().__post_init__()
        config = Config.get()

        self.m_ctx = ZmqContext()
        self.setup_ipc(url=config.app.ipc, ctx=self.m_ctx, type=zmq.DEALER, name='ad_server')


    def handle_ipc(self, buf):
        pass


def create_app(debug):
    app = Starlette(debug=debug)
    ipc = None


    @app.on_event('startup')
    async def startup_handler():
        nonlocal ipc
        ipc = IPCImpl()
        ipc.task = asyncio.create_task(ipc.run_ipc())


    @app.on_event('shutdown')
    async def shutdown_handler():
        nonlocal ipc
        if ipc.task and not ipc.task.done():
            ipc.close_ipc()
            await ipc.task


    @app.route('/cgi/registry.exe')
    class RegisterAd(HTTPEndpoint):
        async def post(self, request):
            nonlocal ipc

            ip = request.client.host
            port = 1001

            # Port and Ip are optional
            try:
                ip = request.query_params['IP']
            except:
                pass

            try:
                port = int(request.query_params['PORT'])
            except:
                pass

            # Name is a must
            try:
                name = request.query_params['NAME']
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            name = name.upper()
            updated, _ = await models.upsert_device(
                name=name,
                address=ip,
                port=port)

            if updated:
                obj = {
                    'cmd': 'notify-datachanged'
                }

                try:
                    buf = json.dumps(obj).encode()
                except:
                    logging.exception('Object to json conversion error')

                await ipc.send_ipc(buf)

            return HTMLResponse('<html><body><strong>\nSuccessfully Registered!\n</strong></body></html>')

    return app


def run_adserver(**kwargs):
    Config.init(board_cfg_file())
    setup_logger('ad_server.txt')


    config = Config.get()
    models.SessionSingleton.recreate(config.app.database)

    app = create_app(True)
    uvicorn.run(app, host=config.pos.ad_host, port=int(config.pos.ad_port))


if __name__ == '__main__':
    run_adserver()
